@@ -15,13 +15,14 @@
 // Copyright (C) 2007 Ilmari Heikkinen <ilmari.heikkinen@gmail.com>
 // Copyright (C) 2009 Shen Liang <shenzhuxi@gmail.com>
 // Copyright (C) 2009 Stefan Thomas <thomas@eload24.com>
-// Copyright (C) 2010, 2012 Adrian Johnson <ajohnson@redneon.com>
+// Copyright (C) 2010, 2012, 2017 Adrian Johnson <ajohnson@redneon.com>
 // Copyright (C) 2010 Harry Roberts <harry.roberts@midnight-labs.org>
 // Copyright (C) 2010 Christian Feuersnger <cfeuersaenger@googlemail.com>
 // Copyright (C) 2010, 2015 William Bader <williambader@hotmail.com>
 // Copyright (C) 2011-2013 Thomas Freitag <Thomas.Freitag@alfa.de>
 // Copyright (C) 2012 Anthony Wesley <awesley@smartnetworks.com.au>
 // Copyright (C) 2015 Adam Reichold <adamreichold@myopera.com>
+// Copyright (C) 2016 Kenji Uno <ku@digitaldolphins.jp>
 //
 // To see a description of the changes please see the Changelog file that
 // came with your tarball or type make ChangeLog if you are building from git
@@ -89,7 +90,7 @@ SplashBitmap::SplashBitmap(int widthA, int heightA, int rowPadA,
       rowSize = -1;
     }
     break;
-#if SPLASH_CMYK
+#ifdef SPLASH_CMYK
   case splashModeCMYK8:
     if (width > 0 && width <= INT_MAX / 4) {
       rowSize = width * 4;
@@ -111,7 +112,7 @@ SplashBitmap::SplashBitmap(int widthA, int heightA, int rowPadA,
     rowSize -= rowSize % rowPad;
   }
   data = (SplashColorPtr)gmallocn_checkoverflow(rowSize, height);
-  if (data != NULL) {
+  if (data != nullptr) {
     if (!topDown) {
       data += (height - 1) * rowSize;
       rowSize = -rowSize;
@@ -119,20 +120,20 @@ SplashBitmap::SplashBitmap(int widthA, int heightA, int rowPadA,
     if (alphaA) {
       alpha = (Guchar *)gmallocn(width, height);
     } else {
-      alpha = NULL;
+      alpha = nullptr;
     }
   } else {
-    alpha = NULL;
+    alpha = nullptr;
   }
   separationList = new GooList();
-  if (separationListA != NULL)
+  if (separationListA != nullptr)
     for (int i = 0; i < separationListA->getLength(); i++)
       separationList->append(((GfxSeparationColorSpace *) separationListA->get(i))->copy());
 }
 
 SplashBitmap *SplashBitmap::copy(SplashBitmap *src) {
   SplashBitmap *result = new SplashBitmap(src->getWidth(), src->getHeight(), src->getRowPad(), 
-    src->getMode(), src->getAlphaPtr() != NULL, src->getRowSize() >= 0, src->getSeparationList());
+    src->getMode(), src->getAlphaPtr() != nullptr, src->getRowSize() >= 0, src->getSeparationList());
   Guchar *dataSource = src->getDataPtr();
   Guchar *dataDest = result->getDataPtr();
   int amount = src->getRowSize();
@@ -144,7 +145,7 @@ SplashBitmap *SplashBitmap::copy(SplashBitmap *src) {
     amount *= src->getHeight();
   }
   memcpy(dataDest, dataSource, amount);
-  if (src->getAlphaPtr() != NULL) {
+  if (src->getAlphaPtr() != nullptr) {
     memcpy(result->getAlphaPtr(), src->getAlphaPtr(), src->getWidth() * src->getHeight());
   }
   return result;
@@ -246,7 +247,7 @@ SplashError SplashBitmap::writePNMFile(FILE *f) {
     }
     break;
 
-#if SPLASH_CMYK
+#ifdef SPLASH_CMYK
   case splashModeCMYK8:
   case splashModeDeviceN8:
     // PNM doesn't support CMYK
@@ -307,7 +308,7 @@ void SplashBitmap::getPixel(int x, int y, SplashColorPtr pixel) {
     pixel[1] = p[1];
     pixel[2] = p[0];
     break;
-#if SPLASH_CMYK
+#ifdef SPLASH_CMYK
   case splashModeCMYK8:
     p = &data[y * rowSize + 4 * x];
     pixel[0] = p[0];
@@ -332,11 +333,11 @@ SplashColorPtr SplashBitmap::takeData() {
   SplashColorPtr data2;
 
   data2 = data;
-  data = NULL;
+  data = nullptr;
   return data2;
 }
 
-SplashError SplashBitmap::writeImgFile(SplashImageFileFormat format, char *fileName, int hDPI, int vDPI, const char *compressionString) {
+SplashError SplashBitmap::writeImgFile(SplashImageFileFormat format, char *fileName, int hDPI, int vDPI, WriteImgParams* params) {
   FILE *f;
   SplashError e;
 
@@ -344,16 +345,29 @@ SplashError SplashBitmap::writeImgFile(SplashImageFileFormat format, char *fileN
     return splashErrOpenFile;
   }
 
-  e = writeImgFile(format, f, hDPI, vDPI, compressionString);
-  
+  e = writeImgFile(format, f, hDPI, vDPI, params);
+
   fclose(f);
   return e;
 }
 
-SplashError SplashBitmap::writeImgFile(SplashImageFileFormat format, FILE *f, int hDPI, int vDPI, const char *compressionString) {
+void SplashBitmap::setJpegParams(ImgWriter *writer, WriteImgParams* params)
+{
+#ifdef ENABLE_LIBJPEG
+  if (params) {
+    static_cast<JpegWriter*>(writer)->setProgressive(params->jpegProgressive);
+    if (params->jpegQuality >= 0)
+      static_cast<JpegWriter*>(writer)->setQuality(params->jpegQuality);
+  }
+#endif
+}
+
+SplashError SplashBitmap::writeImgFile(SplashImageFileFormat format, FILE *f, int hDPI, int vDPI, WriteImgParams* params) {
   ImgWriter *writer;
 	SplashError e;
   
+  SplashColorMode imageWriterFormat = splashModeRGB8;
+
   switch (format) {
     #ifdef ENABLE_LIBPNG
     case splashFormatPng:
@@ -362,13 +376,15 @@ SplashError SplashBitmap::writeImgFile(SplashImageFileFormat format, FILE *f, in
     #endif
 
     #ifdef ENABLE_LIBJPEG
-    #if SPLASH_CMYK
+    #ifdef SPLASH_CMYK
     case splashFormatJpegCMYK:
       writer = new JpegWriter(JpegWriter::CMYK);
+      setJpegParams(writer, params);
       break;
     #endif
     case splashFormatJpeg:
       writer = new JpegWriter();
+      setJpegParams(writer, params);
       break;
     #endif
 	
@@ -377,15 +393,17 @@ SplashError SplashBitmap::writeImgFile(SplashImageFileFormat format, FILE *f, in
       switch (mode) {
       case splashModeMono1:
         writer = new TiffWriter(TiffWriter::MONOCHROME);
+        imageWriterFormat = splashModeMono1;
         break;
       case splashModeMono8:
         writer = new TiffWriter(TiffWriter::GRAY);
+        imageWriterFormat = splashModeMono8;
         break;
       case splashModeRGB8:
       case splashModeBGR8:
         writer = new TiffWriter(TiffWriter::RGB);
         break;
-#if SPLASH_CMYK
+#ifdef SPLASH_CMYK
       case splashModeCMYK8:
       case splashModeDeviceN8:
         writer = new TiffWriter(TiffWriter::CMYK);
@@ -395,8 +413,8 @@ SplashError SplashBitmap::writeImgFile(SplashImageFileFormat format, FILE *f, in
         fprintf(stderr, "TiffWriter: Mode %d not supported\n", mode);
         writer = new TiffWriter();
       }
-      if (writer) {
-        ((TiffWriter *)writer)->setCompressionString(compressionString);
+      if (writer && params) {
+        ((TiffWriter *)writer)->setCompressionString(params->tiffCompression.getCString());
       }
       break;
     #endif
@@ -408,7 +426,7 @@ SplashError SplashBitmap::writeImgFile(SplashImageFileFormat format, FILE *f, in
       return splashErrGeneric;
   }
 
-	e = writeImgFile(writer, f, hDPI, vDPI);
+	e = writeImgFile(writer, f, hDPI, vDPI, imageWriterFormat);
 	delete writer;
 	return e;
 }
@@ -425,7 +443,7 @@ void SplashBitmap::getRGBLine(int yl, SplashColorPtr line) {
     m = byteToDbl(col[1]);
     y = byteToDbl(col[2]);
     k = byteToDbl(col[3]);
-#if SPLASH_CMYK
+#ifdef SPLASH_CMYK
     if (separationList->getLength() > 0) {
       for (int i = 0; i < separationList->getLength(); i++) {
         if (col[i+4] > 0) {
@@ -461,7 +479,7 @@ void SplashBitmap::getRGBLine(int yl, SplashColorPtr line) {
   }
 }
 
-void SplashBitmap::getXBGRLine(int yl, SplashColorPtr line, bool useAlpha) {
+void SplashBitmap::getXBGRLine(int yl, SplashColorPtr line, ConversionMode conversionMode) {
   SplashColor col;
   double c, m, y, k, c1, m1, y1, k1, r, g, b;
 
@@ -471,7 +489,7 @@ void SplashBitmap::getXBGRLine(int yl, SplashColorPtr line, bool useAlpha) {
     m = byteToDbl(col[1]);
     y = byteToDbl(col[2]);
     k = byteToDbl(col[3]);
-#if SPLASH_CMYK
+#ifdef SPLASH_CMYK
     if (separationList->getLength() > 0) {
       for (int i = 0; i < separationList->getLength(); i++) {
         if (col[i+4] > 0) {
@@ -501,16 +519,34 @@ void SplashBitmap::getXBGRLine(int yl, SplashColorPtr line, bool useAlpha) {
     y1 = 1 - y;
     k1 = 1 - k;
     cmykToRGBMatrixMultiplication(c, m, y, k, c1, m1, y1, k1, r, g, b);
-    *line++ = dblToByte(clip01(b));
-    *line++ = dblToByte(clip01(g));
-    *line++ = dblToByte(clip01(r));
-    *line++ = useAlpha ? getAlpha(x, yl) : 255;
+
+    if (conversionMode == conversionAlphaPremultiplied) {
+        const double a = getAlpha(x, yl) / 255.0;
+
+        *line++ = dblToByte(clip01(b * a));
+        *line++ = dblToByte(clip01(g * a));
+        *line++ = dblToByte(clip01(r * a));
+    } else {
+        *line++ = dblToByte(clip01(b));
+        *line++ = dblToByte(clip01(g));
+        *line++ = dblToByte(clip01(r));
+    }
+
+    if (conversionMode != conversionOpaque) {
+        *line++ = getAlpha(x, yl);
+    } else {
+        *line++ = 255;
+    }
   }
 }
 
-GBool SplashBitmap::convertToXBGR(bool useAlpha) {
+static inline Guchar div255(int x) {
+  return (Guchar)((x + (x >> 8) + 0x80) >> 8);
+}
+
+GBool SplashBitmap::convertToXBGR(ConversionMode conversionMode) {
   if (mode == splashModeXBGR8) {
-    if (useAlpha) {
+    if (conversionMode != conversionOpaque) {
       // Copy the alpha channel into the fourth component so that XBGR becomes ABGR.
       const SplashColorPtr dbegin = data;
       const SplashColorPtr dend = data + rowSize * height;
@@ -518,11 +554,20 @@ GBool SplashBitmap::convertToXBGR(bool useAlpha) {
       Guchar *const abegin = alpha;
       Guchar *const aend = alpha + width * height;
 
-      SplashColorPtr d = dbegin + 3;
+      SplashColorPtr d = dbegin;
       Guchar *a = abegin;
 
-      for(; d < dend && a < aend; d += 4, a += 1) {
-        *d = *a;
+      if (conversionMode == conversionAlphaPremultiplied) {
+          for (; d < dend && a < aend; d += 4, a += 1) {
+              d[0] = div255(d[0] * *a);
+              d[1] = div255(d[1] * *a);
+              d[2] = div255(d[2] * *a);
+              d[3] = *a;
+          }
+      } else {
+          for (d += 3; d < dend && a < aend; d += 4, a += 1) {
+              *d = *a;
+          }
       }
     }
 
@@ -531,10 +576,10 @@ GBool SplashBitmap::convertToXBGR(bool useAlpha) {
   
   int newrowSize = width * 4;
   SplashColorPtr newdata = (SplashColorPtr)gmallocn_checkoverflow(newrowSize, height);
-  if (newdata != NULL) {
+  if (newdata != nullptr) {
     for (int y = 0; y < height; y++) {
       unsigned char *row = newdata + y * newrowSize;
-      getXBGRLine(y, row, useAlpha);
+      getXBGRLine(y, row, conversionMode);
     }
     if (rowSize < 0) {
       gfree(data + (height - 1) * rowSize);
@@ -545,10 +590,10 @@ GBool SplashBitmap::convertToXBGR(bool useAlpha) {
     rowSize = newrowSize;
     mode = splashModeXBGR8;
   }
-  return newdata != NULL;
+  return newdata != nullptr;
 }
 
-#if SPLASH_CMYK
+#ifdef SPLASH_CMYK
 void SplashBitmap::getCMYKLine(int yl, SplashColorPtr line) {
   SplashColor col;
 
@@ -590,9 +635,9 @@ void SplashBitmap::getCMYKLine(int yl, SplashColorPtr line) {
 }
 #endif
 
-SplashError SplashBitmap::writeImgFile(ImgWriter *writer, FILE *f, int hDPI, int vDPI) {
+SplashError SplashBitmap::writeImgFile(ImgWriter *writer, FILE *f, int hDPI, int vDPI, SplashColorMode imageWriterFormat) {
   if (mode != splashModeRGB8 && mode != splashModeMono8 && mode != splashModeMono1 && mode != splashModeXBGR8 && mode != splashModeBGR8
-#if SPLASH_CMYK
+#ifdef SPLASH_CMYK
       && mode != splashModeCMYK8 && mode != splashModeDeviceN8
 #endif
      ) {
@@ -605,7 +650,7 @@ SplashError SplashBitmap::writeImgFile(ImgWriter *writer, FILE *f, int hDPI, int
   }
 
   switch (mode) {
-#if SPLASH_CMYK
+#ifdef SPLASH_CMYK
     case splashModeCMYK8:
       if (writer->supportCMYK()) {
         SplashColorPtr row;
@@ -717,41 +762,81 @@ SplashError SplashBitmap::writeImgFile(ImgWriter *writer, FILE *f, int hDPI, int
     
     case splashModeMono8:
     {
-      unsigned char *row = new unsigned char[3 * width];
-      for (int y = 0; y < height; y++) {
-        // Convert into a PNG row
-        for (int x = 0; x < width; x++) {
-          row[3*x] = data[y * rowSize + x];
-          row[3*x+1] = data[y * rowSize + x];
-          row[3*x+2] = data[y * rowSize + x];
-        }
+      if (imageWriterFormat == splashModeMono8) {
+        SplashColorPtr row;
+        unsigned char **row_pointers = new unsigned char*[height];
+        row = data;
 
-        if (!writer->writeRow(&row)) {
-          delete[] row;
+        for (int y = 0; y < height; ++y) {
+          row_pointers[y] = row;
+          row += rowSize;
+        }
+        if (!writer->writePointers(row_pointers, height)) {
+          delete[] row_pointers;
           return splashErrGeneric;
         }
+        delete[] row_pointers;
+      } else if (imageWriterFormat == splashModeRGB8) {
+        unsigned char *row = new unsigned char[3 * width];
+        for (int y = 0; y < height; y++) {
+          // Convert into a PNG row
+          for (int x = 0; x < width; x++) {
+            row[3*x] = data[y * rowSize + x];
+            row[3*x+1] = data[y * rowSize + x];
+            row[3*x+2] = data[y * rowSize + x];
+          }
+
+          if (!writer->writeRow(&row)) {
+            delete[] row;
+            return splashErrGeneric;
+          }
+        }
+        delete[] row;
+      }
+      else {
+        // only splashModeMono8 or splashModeRGB8
+        return splashErrGeneric;
       }
-      delete[] row;
     }
     break;
     
     case splashModeMono1:
     {
-      unsigned char *row = new unsigned char[3 * width];
-      for (int y = 0; y < height; y++) {
-        // Convert into a PNG row
-        for (int x = 0; x < width; x++) {
-          getPixel(x, y, &row[3*x]);
-          row[3*x+1] = row[3*x];
-          row[3*x+2] = row[3*x];
-        }
+      if (imageWriterFormat == splashModeMono1) {
+        SplashColorPtr row;
+        unsigned char **row_pointers = new unsigned char*[height];
+        row = data;
 
-        if (!writer->writeRow(&row)) {
-          delete[] row;
+        for (int y = 0; y < height; ++y) {
+          row_pointers[y] = row;
+          row += rowSize;
+        }
+        if (!writer->writePointers(row_pointers, height)) {
+          delete[] row_pointers;
           return splashErrGeneric;
         }
+        delete[] row_pointers;
+      } else if (imageWriterFormat == splashModeRGB8) {
+        unsigned char *row = new unsigned char[3 * width];
+        for (int y = 0; y < height; y++) {
+          // Convert into a PNG row
+          for (int x = 0; x < width; x++) {
+            getPixel(x, y, &row[3*x]);
+            row[3*x+1] = row[3*x];
+            row[3*x+2] = row[3*x];
+          }
+
+          if (!writer->writeRow(&row)) {
+            delete[] row;
+            return splashErrGeneric;
+          }
+        }
+        delete[] row;
+      }
+      else {
+        // only splashModeMono1 or splashModeRGB8
+        return splashErrGeneric;
       }
-      delete[] row;
     }
     break;
     
